"""
zwiki mailin - post an incoming email message to a zwiki page

This is intended to be set up as an external method for receiving mail
from a mailer alias, procmail recipe or script via something like
| curl -n -F 'msg=<-' http://mysite/mywikifolder/mailin
The mailin method expects at least one argument, an RFC822 message,
which will be formatted as a wiki comment and appended to the
appropriate page.  Here are the default delivery rules:

- if called in a page context, always use that page. Otherwise,

- if a recipient of the form
  "[something](wiki|tracker|bugs|mailin)@virtualhost" is found where
  virtualhost matches an existing VirtualHostMonster entry, then in the
  corresponding folder, otherwise in the current folder, use:

   (disabled) 1. a wiki page name before @ in a recipient header
   2. or the first wiki page name appearing in the subject
   3. or the folder's default_page property
   4. or the supplied defaultpage argument
   5. or the hardcoded DEFAULTPAGE below

- if a non-existent page is named in the subject, create it, and add a
  link on the default page

- if the trackerissue flag is passed or any of the recipients contains
  "(tracker)" or there was a match for
  "[something](tracker|bugs)@virtualhost", create a tracker issue page
  instead of posting a comment

- unless subscribersonly=0 or the folder's posting_policy property is
  'open': if the sender is not subscribed anywhere in the wiki, bounce the
  message

- if the message appears to have been mailed out by zwiki, silently
  discard it

- if the message appears to be an auto-response, silently discard it

todo:
friendly bounce messages
size limits
"""

from types import *
import re, string 
import rfc822, mimetools, multifile, StringIO
from Products.ZWiki.Regexps import wikiname1,wikiname2,bracketedexpr

DEFAULTPAGE = 'FrontPage'
#WIKINAMEEXP = r'(%s|%s|%s)' % (wikiname1,wikiname2,bracketedexpr)
WIKINAMEEXP = r'(%s|%s)' % (wikiname1,wikiname2)
MAILINADDREXP = r'(wiki|tracker|bugs|mailin)@'
TRACKERADDREXP = r'(tracker|bugs)$'

class IncomingMessage:
    """
    an incoming wikimail message

    responsibilities: parse an rfc822 and expose the parts of interest
    """
    def __init__(self,message):
        self.raw = message
        self.stream = StringIO.StringIO(self.raw)
        self.msg = mimetools.Message(self.stream)
        self.date = self.msg['Date']
        self.subject = (self.msg.has_key('Subject') and self.msg['Subject']) or ''
        self.sender = self.msg['From']
        self.senderRealName = rfc822.parseaddr(self.sender)[0]
        self.senderEmail = rfc822.parseaddr(self.sender)[1]
        self.senderUserName = (self.senderRealName or
                               re.sub(r'@.*$',r'',self.senderEmail))
        self.recipients = []
        for hdr in ['To','Cc','Bcc']:
            if self.msg.has_key(hdr):
                self.recipients.extend(self.msg.getaddrlist(hdr))
        self.recipients = map(lambda x:x[1], self.recipients)
        self.wikiRecipients = filter(lambda x:re.search(MAILINADDREXP,x),
                                     self.recipients)
        self.wikiRecipients = map(lambda x:string.split(x,'@'),
                                  self.wikiRecipients)
        if self.isMime():
            self.body = self.firstMimePartMatching('text/plain')
        else:
            self.body = self.msg.fp.read()
        # strip trailing newlines that seem to get added in transit
        self.body = re.sub(r'(?s)\n+$',r'\n',self.body)

    def isMime(self):
        return re.search(r'Content-Type:.*multipart',self.raw)

    # based on python manual example
    def firstMimePartMatching(self,mimetype):
        msgtype = self.msg.gettype()
        params = self.msg.getplist()
        data = StringIO.StringIO()
        if msgtype[:10] == "multipart/":
            file = multifile.MultiFile(self.stream)
            file.push(self.msg.getparam("boundary"))
            while file.next():
                submsg = mimetools.Message(file)
                try:
                    data = StringIO.StringIO()
                    mimetools.decode(file, data, submsg.getencoding())
                except ValueError:
                    continue
                if submsg.gettype() == mimetype:
                    break
            file.pop()
        return data.getvalue()

    
def findVirtualHostFor(self, msg):
    """
    If we have one of the special recipients listed above along with a
    matching virtual host, return the virtual host's folder; otherwise None.
    Also save the username part of the matched recipient for later.
    """
    folder = None
    vurt = getattr(self,'vurt',None)
    if vurt and msg.wikiRecipients:
        for alias, host in msg.wikiRecipients:
            # based on SiteAccess
            dict = getattr(vurt, 'domain_path', None)
            if dict:
                real_host=None
                l_host = string.lower(host)
                if vurt.domain_path.has_key(l_host):
                    real_host=l_host
                else:
                    for hostname in vurt.domain_path.keys():
                        if hostname[:2] == '*.':
                            if string.find(l_host, hostname[2:]) != -1:
                                real_host = hostname
                if real_host:
                    path = vurt.domain_path[real_host]
                    folder = self.getPhysicalRoot().unrestrictedTraverse(path)
                    msg.aliasUsed = alias
                    break
    return folder

def findDestinationFor(self, msg, pagenameexp, defaultpage,
                       checkrecipient, checksubject):
    """
    figure out the destination wiki folder and page for an incoming
    message. See rules above.
    """
    if self.meta_type == 'ZWiki Page':
        folder, page, pagename, creating = self.aq_parent, self, self.id(), 0
    else:
        folder = findVirtualHostFor(self,msg) or self
        defaultpagename = getattr(folder,'default_page',defaultpage)
        pagename = None
        #if checkrecipient:
        #    pagename = re.search(pagenameexp, string.join(recipients))
        #    if pagename:
        #        pagename = pagename.group(1)
        # or in the subject
        if (not pagename) and checksubject:
            pagename = re.search(pagenameexp,msg.subject)
            if pagename:
                pagename = pagename.group(1)
        if not pagename:
            pagename = defaultpagename

        if hasattr(folder,pagename):
            page = getattr(folder,pagename)
            creating = 0
        else:
            page = getattr(folder,defaultpagename)
            creating = 1

    return (folder,page,pagename,creating)

def isJunk(msg):
    """
    Return true if this message should be silently ignored.

    Ideally, this should block mail loops, auto-responders and spam,
    but allow mailing list messages and mailouts from other zwikis.
    Currently it flags everything that looks like it came from a bot.
    Actually I don't think works as a way to filter autoresponders.
    We should make sure to include mailing headers in our mailouts
    and then most autoresponders should ignore us.

    qmail-autoresponder's bot-filtering procedure is reportedly good - see
    http://untroubled.org/qmail-autoresponder/procedure.txt .
    TMDA and spamassassin are two good spam filters - see
    http://software.libertine.org/tmda ,
    http://spamassassin.taint.org .
    """
    # a zwiki mailout
    if re.search(r'(?mi)^X-Zwiki-Version:',msg):
        return 1
    # the most common auto-response subject
    if re.search(r'(?mi)^Subject:.*out of office',msg):
        return 1
    # mailing list or low precedence mail
    # XXX should allow these, but need to pass through the list loop
    # headers to avoid mail loop between mutually-subscribed zwiki and
    # mail list
    if re.search(r'(?mi)^(List-ID|(X-)?Mailing-List|X-ML-Name):',msg):
        return 1
    if re.search(
        r'(?mi)^List-(Help|Unsubscribe|Subscribe|Post|Owner|Archive):',msg):
        return 1
    if re.search(r'(?mi)^Precedence:\s*(junk|bulk|list)\s*$',msg):
        return 1
    return 0

def mailin(self, msg, pagenameexp=WIKINAMEEXP, defaultpage=DEFAULTPAGE,
           separator='\n\n', checkrecipient=0, checksubject=1, trackerissue=0,
           subscribersonly=1):

    if isJunk(msg): return

    msg = IncomingMessage(msg)
    #XXX make this a method later
    folder,page,pagename,creating = findDestinationFor(self,msg,
                                              pagenameexp,defaultpage,
                                              checkrecipient,checksubject)
    if (subscribersonly and
        not getattr(folder,'posting_policy',None) == 'open'):
        if not len(page.allSubscriptionsFor(msg.senderEmail)):
            return '\nSorry, you must be a subscriber to send mail to this wiki.\n\n\n'

    # stash the mailin username into REQUEST so zwiki_username_or_ip
    # will use it to set last_editor
    self.REQUEST.set('MAILIN_USERNAME', msg.senderUserName)

    # create a tracker issue ?
    # somewhat messy and temporary: check if the trackerissue flag is set,
    # or the mailin alias used looks like a tracker alias,
    # or (last resort) one of the recipients contains "(tracker)"
    if (trackerissue or
        re.search(TRACKERADDREXP,getattr(msg,'aliasUsed','')) or
        re.search(r'\(tracker\)', string.join(msg.recipients))):
        subject = msg.subject or 'no subject'
        self.REQUEST.set('newtitle', subject)
        # italicize quoted text in replies (as in comment(), but remember DRY)
        body = re.sub(r'(?m)^>(.*)',r'<br>><i>\1</i>',msg.body)
        self.REQUEST.set('newtext', body)
        self.REQUEST.set('submitted', 1)
        page.aq_parent.ZwikiTracker(REQUEST=self.REQUEST)
        return 

    # create a page ?
    if creating:
        page.comment(use_heading=1,
                     text=pagename+' created',
                     username=msg.senderUserName,
                     note=' (new page)',
                     REQUEST=self.REQUEST)
        # zwiki.org: REQUEST has no URL2, create fails. Leave out for now
        page.create(pagename,text='') #,REQUEST=self.REQUEST) 
        page = getattr(self,pagename)

    # add comment
    # use the time of sending, or
    #page.comment(use_heading=1, text=text, username=sender, time=date, note=' (via mail)')
    # use the time of posting to the wiki - see how this works out
    page.comment(use_heading=1, text=msg.body,
                 username=msg.senderUserName,
                 note=' (via mail)',REQUEST=self.REQUEST)
#    page.comment(use_heading=1, text=str(self.REQUEST),
#                 username=msg.senderUserName,
#                 note=' (debug)',REQUEST=self.REQUEST)
